#!/usr/bin/perl -w
use strict;

# Copyright (c) 2001,2002 Ognyan Kulev <ogi@fmi.uni-sofia.bg>.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The GNU General Public License text can be found at
# http://www.gnu.org/copyleft/gpl.txt
#
# ChangeLog:
#
# 2001-06-26
#       Initial version 1.0.  Supports plain text, HTML and
#       one HTML page per question (fancy html).
# 2001-12-19
#       Fix: Convert `&' to `&amp;' in [X]HTML outputs.
#       Recognize `text' and ``text''.
# 2002-01-19
#       Added documentation for `text' and ``text'' cases.
#       Documentation fixes and much more comments.
# 2002-01-20
#       Values of directives are now able to contain character `}'
#       by using escaping.
# 2002-01-21
#       Nested `text' and ``text'' now work as expected in all cases
#       (I hope :).
# 2002-01-22
#       Generate HTML 4.01 Strict or XHTML 1.0 Strict conformant
#       output depending on requiring html or xhtml respectively.
#       Adding directives that can change all English texts that
#       are generated by the program.  Suggested by Lars Weber
#       <me@lars.in-berlin.de>.
# 2002-01-23
#       Announced that this is version 2.0.
# 2002-02-06
#       The Table Of Content and the headings in HTML output are now GNUish.
#       Added all long options.  This broke the program.
#       In POD the `C<text>' style is replaced with C<text> style
#       because perl 5.6 automatically puts surrounding quotes in text
#       output and the text is visibly different in HTML output.
# 2002-02-12
#	The program is now workable again.  It can use template too.
# 2002-02-19
#	Add --single-quotes.  --single-quotes and --double-quotes now work.
#	(There was only documentation for them before but no code.)
# 2002-02-20
#	Fancy html is synchronized with the previous changes.
#	Announced that this is version 3.0.
# 2002-02-27
#	As noted by Marcus Brinkmann <Marcus.Brinkmann@ruhr-uni-bochum.de>
#	text output didn't work.  This is fixed along with some tiny bugs.
# 2002-03-03
#	Fix fancy html output bugs.  It is ensured that html outputs
#	are valid HTML (http://validator.w3.org).  Unfortunately
#	the NAME attribute of A elements is invalid in HTML 4.01 Strict.
# 2002-03-06
#	Add --strict-html.  CGI::Pretty is not used anymore because
#	it breaks XHTML output by using upper-case element names.
# 2002-03-09
#	Minor changes.
# 2002-03-21
#	More minor changes.
# 2002-03-24
#	Removed some newline prints in fancy html mode.  Reported by
#	Lars Weber <me@lars.in-berlin.de>.

=head1 NAME

gen-faq.pl - converting glibc-style FAQ to plain text or HTML

=head1 SYNOPSIS

 gen-faq.pl [--help] [--version]
            [--output-type=TYPE] [-t TYPE]
            [--output=FILENAME] [-o FILENAME]
            [--html-template=FILE] [-m FILE]
            [--[no]xhtml] [--[no]strict-html] [--[no]intro-in-toc]
            [-1 FORMAT] [--single-quotes=FORMAT]
            [-2 FORMAT] [--double-quotes=FORMAT]
            FILENAME

=head1 DESCRIPTION

This is a program to convert FAQ to a plain text file, one html file
or a web of html pages.  Optionally it can generate XHTML 1.0 text
instead of HTML 4.01 text.

Options control how the output will look like.  Directives in the FAQ
set properties of content.

=head1 OPTIONS

=over 4

=item C<--help>

=item C<--version>

Show program version and the possible options.

=item C<--output-type=TYPE>

=item C<-t TYPE>

Set the desired output format.  It can be C<text> (the default),
C<html> or C<fancy-html>.  C<text> and C<html> generate one output
file that, by default, is put in file with the name of the FAQ file
name and suffix C<.txt> or C<.html> added, respectively.
C<fancy-html> generates many files, one for each question and each
section, that by default are put in directory with the name of the FAQ
file name and suffix C<.d> added.

=item C<--output=FILENAME>

=item C<-o FILENAME>

If plain text file or single html file is generated then FILENAME is
the name of that output file.  If fancy html output is generated then
FILENAME is the directory where the output files has to be put in.

=item C<--html-template=FILENAME>

=item C<-m FILENAME>

I<In html and fancy html modes.> The file name of the template file for
the single output file (in html mode) or for all HTML files (in fancy
html mode).  Each output file is the same as this template except that
the following strings are replaced:

=over 4

=item C<E<lt>gen-faq:charset/E<gt>>

The value set by directive C<?{charset}>.

=item C<E<lt>gen-faq:title/E<gt>>

The title of the FAQ.  That is, the first paragraph.

=item C<E<lt>gen-faq:all/E<gt>>

The Table of Content and then the content itself.

=back

=item C<--[no]xhtml>

I<In html and fancy html modes only.> Whether the output must be XHTML
confomant.  Default is C<--noxhtml>.

=item C<--[no]strict-html>

I<In html mode and fancy html modes.> Whether the output must be HTML
4 Strict conformant or HTML 4 Transitional conformant.  Currently the
only difference is whether ID attribute or NAME attribute is used in
destination anchor elements and this option doesn't affect fancy html
mode because it doesn't use destination anchors.  ID attribute is used
when C<--strict-html> is requested and NAME attribute is used when
C<--nostrict-html> is requested.  Default is C<--nostrict-html>.

=item C<--[no]intro-in-toc>

I<In html mode only.> Whether the introduction paragraphs have to be
after the Table Of Content as sole section.  This is GNU style for
html.  Default is C<--nointro-in-toc>.

=item C<--single-quotes=FORMAT>

=item C<-1 FORMAT>

I<In html and fancy html modes only.> Controls how C<`text'> style
markup should be formatted in html outputs.  Possible values are:

=over 4

=item C<q>

C<`text'>.  Don't recognize this markup.

=item C<tt>

C<E<lt>ttE<gt>textE<lt>/ttE<gt>>.  Make text to be monospace and strip
quotes.

=item C<qtt>

C<`E<lt>ttE<gt>textE<lt>/ttE<gt>'>.  This is the default formatting.
Make text to be monospace and leave quotes.

=back

=item C<--double-quotes=FORMAT>

=item C<-2 FORMAT>

I<In html and fancy html modes only.> Controls how C<``text''> style
markup should be formatted in html outputs.  Possible values are:

=over 4

=item C<q>

C<``text''>.  This is the default formatting.  In other words, don't
recognize this markup.

=item C<em>

C<E<lt>emE<gt>textE<lt>/emE<gt>>.  Emphasize text and strip quotes.

=item C<qem>

C<``E<lt>emE<gt>textE<lt>/emE<gt>''>.  Emphasize text and leave
quotes.

=back

=back

=head1 FAQ FORMAT

=head2 Paragraphs

The whole text is a sequence of I<paragraphs>.  Each paragraph is
separated from others by empty line(s).  The only exception of this
rule is that line starting with C<*> or C<-> followed by space mark a
new paragraph even if the preceding line wasn't empty.  (These are list
item paragraphs.)

Remember: your FAQ must look good in its source file because no
formatting will be applied when text output is generated.

=head2 Directives

Before text is split into paragraphs, all directives C<?{NAME=VALUE}>
are processed.  These directives are not shown in any output.  They
must be the first thing in line and not preceded by any spaces.  The
recommended place for directives is after the third C<^L> character in
FAQ file (see section Structure below).  Escape character is C<\> so
all C<}> characters in VALUE must be replaced with C<\}> and all C<\>
characters must be replaced with C<\\>.  A directive must fit in one
line.  Using HTML markup in directives for text is not recommended.

=over

=item C<?{text-toc=TEXT}>

I<In all modes.> The text to be used instead of I<Table Of
Contents>.

=item C<?{text-q=TEXT}>

I<In all modes.> The text to be used instead of I<Q> when
generating references to questions.

=item C<?{charset=VALUE}>

I<In html and fancy html modes.> This is the charset that is stored in
the META element of all html.

=item C<?{title-prefix=TEXT}>

I<In fancy html mode only.> Prefix for titles of the pages in fancy
html output.

=item C<?{text-sections=TEXT}>

I<In fancy html mode only.> The text to be used instead of
I<Sections>.

=item C<?{text-this-section=TEXT}>

I<In fancy html mode only.> The text to be used instead of I<This
section>.

=item C<?{text-questions=TEXT}>

I<In fancy html mode only.> The text to be used instead of
I<Questions>.

=item C<?{text-next=TEXT}>

I<In fancy html mode only.> The text to be used instead of
I<Next>.

=item C<?{text-prev=TEXT}>

I<In fancy html mode only.> The text to be used instead of
I<Previous>.

=item C<?{text-all=TEXT}>

I<In fancy html mode only.> The text to be used instead of I<All
sections>.

=back

=head2 Structure

As a second special syntax is the usage of C<^L> character.  It is
used to separate the four parts of the FAQ.  A FAQ is expected to look
like this:

    TITLE (the first paragraph)
    TOP TEXT (what is this FAQ for, etc)
    ^L
    QUESTIONS
    ^L
    BOTTOM TEXT (who contribute, etc)
    ^L
    (the rest of the text is ignored but directives are processed)

=head2 Sections and questions

I<Questions> are grouped into I<sections>.  Paragraph that starts with
C<?> followed by space is considered start of a new section.  (The TAB
character is considered space too.)  The content of this paragraph is
the name of the section.  Paragraph that starts with C<??> followed by
space is considered start of a new question.  Optionally immediately
after the string C<??> you can put identifier (sequence of
alpha-numeric characters) that can be used elsewhere to refer to this
question.  The rest of the lines of this question must start with
exactly one tab.  This is needed for better plain text output.  After
the question all ordinary paragraphs are considered as answer for this
question.

=head2 Paragraphs

They can contain hyperlinks to http and ftp resources.  E-mails are
recognized too.  Each occurence of C<?IDENTIFIER> is replaced with
link to the question with this IDENTIFIER.

C<`TEXT'> and C<``TEXT''> are used for monospace text (like user
commands) and for emphasized text (like citations, rendered as italic)
respectively.  These have no effect on text output.  If nesting is
necessary only one level is allowed.  For example, the text
C<```Emphasized'' monospace text'> will be parsed correctly.

=head2 Verbatim paragraphs

Each line of them must start with at least 3 spaces or one tab.  This
is useful for including fragments of programs or user commands.

=head2 List paragraphs

Making lists is easy.  Every list item is a paragraph that starts with
C<-> or C<*> followed by space.  You can write list items without
blank lines between them.  If you want to have a list item with many
paragraphs, each line of the paragraphs must be prefixed by 2 spaces.
If some paragraph is prefixed by more than 2 spaces it will be
considered verbatim paragraph.  If a line doesn't start with 2 spaces
it is recognized as a paragraph out of the list.  You can use verbatim
paragraphs too.  All is summarized in the following example:

        |* First item
        |* Second item.
        |  Second line of the second item.
        |
        |  Second paragraph of the second item.
        |
        |       Verbatim paragraph of the second item.
        |* Third item.
        |
        |       Verbatim paragraph of the third item.
        |
        |Paragraph out of the list.

=head1 HISTORY

This program is a rewrite by Ognyan Kulev
E<lt>C<http://debian.fmi.uni-sofia.bg/~ogi/gen-faq/>E<gt> of a similar
program by Andreas Jaeger E<lt>C<http://www.suse.de/~aj/>E<gt> that
was used to convert GNU C Library FAQ to HTML.  His program was a hack
of a program in the GNU C Library that converts glibc FAQ to plain
text with Table Of Content.

For full ChangeLog see the source.

This program is licensed under GPL version 2 or any later version (at
your opinion).

=cut


# The program has two phases: loading the FAQ into its internal data
# structures and writing the internal data structures in the chosen
# output format.  Here will be described the first phase.

# Before any other processing on a line to take place, all directives
# `?{name=content}' are found, processed and removed.  The content of
# the directives is stored in dedicated variables: $charset and
# $title_prefix, etc.

# Then the text of the FAQ is split into parts by three ^L: top,
# questions, bottom and ignored part.  All parts are processed by one
# engine that parses the paragraphs and puts each of them in the
# @content array.  The content of the first paragraph is stored in
# $title but not in @content.  Each element in this array is one
# paragraph that starts with a meta tag ending with `:'.  All spaces
# and line breaks in paragraphs are preserved.

# Meta tags
# ---------
# normal:
#       Ordinary text paragraph.
# verbatim:
#       Monospace text paragraph.  In HTML outputs it will be written
#       in <PRE> element.
# item-start:
#       The first paragraph of an item in a list.  Notice that
#       the `* ' or `- ' prefix is preserved in the content.
# item-normal:
#       Ordinary text paragraph that is part of a list item.
# item-verbatim:
#       Monospace text paragraph that is part of a list item.
# bottom:
#       Has no content.  It marks the beginning of the third part
#       of the FAQ.  Notice that there is no corresponding paragraph
#       in the FAQ to this meta tag and that the second part of the
#       FAQ is marked by `section:' meta tag, so there is no need for
#       dedicated meta tag for marking the second part of the FAQ.
#       After reaching the fourth part of the text all the rest
#       of the FAQ is ignored but directives are still processed.
# section SectionNumber:
#       Marks section start.  The content of this element is the
#       name of the section.
# question SectionNumber.QuestionNumber:
#       Marks question start.  The content of this element is the
#       name of the question.  If the question was named with an
#       identifier, then this identifier $id is removed from the
#       content of this element and %ref{$id} become
#       `SectionNumber.QuestionNumber'.

# The text in @content elements can have markup too.  The general
# format of these markups is `{NAME=CONTENT}'.

# Original text     Markup text
# -------------     -----------
# \                 \\
# {                 \{
# }                 \}
# ?keyword          {ref=keyword}
# name@domain       {mail=@domain}
# http://...        {href=http://...}
# ftp://...         {href=ftp://...}
# ``text''          ``{tt=text}''
# `text'            `{q=text}'


my $TEXT_WIDTH             = 72;
my $RE_ELEMENT_CONTENT     = "[^\\\}\\\\]*(?:\\\\.[^\\\}\\\\]*)*";
my $RE_ELEMENT_CONTENT_2   = "[^\\\{\\\}\\\\]*(?:(?:\\\\.|\\{$RE_ELEMENT_CONTENT\\})[^\\\{\\\}\\\\]*)*";

########################################################################
# Command options and arguments

use Getopt::Long;
use Pod::Usage;

sub usage {

    my ($message) = @_;
    pod2usage (-message => $message,
               -verbose => 0,
               -exitval => 1);
}

# The default values of the options
my $option_help = 0;
my $option_version = 0;
my $option_output_type = "text";
my $option_output = undef;
my $option_html_template = "";
my $option_xhtml = 0;
my $option_intro_in_toc = 0;
my $option_strict_html = 0;
my $option_single_quotes = "qtt";
my $option_double_quotes = "q";

# Get these options
GetOptions (
        'help' => \$option_help,
        'version' => \$option_version,
        'output-type|t=s' => \$option_output_type,
        'output|o=s' => \$option_output,
        'html-template|m=s' => \$option_html_template,
        'xhtml!' => \$option_xhtml,
        'intro-in-toc!' => \$option_intro_in_toc,
	'strict-html!' => \$option_strict_html,
        'single-quotes|1=s' => \$option_single_quotes,
        'double-quotes|2=s' => \$option_double_quotes
       );

# --help or --version given?
if ($option_help or $option_version or $#ARGV != 0) {
    &usage ('gen-faq.pl - '.
            'converting glibc-style FAQ to plain text or HTML,'.
            ' version 3.0');
}

# Check $option_double_quotes
my $SQUOTES_Q = "q";
my $SQUOTES_TT = "tt";
my $SQUOTES_QTT = "qtt";
$_ = $option_single_quotes;
if ($_ ne $SQUOTES_Q and $_ ne $SQUOTES_TT and $_ ne $SQUOTES_QTT) {
    &usage ("Invalid --single-quotes=$option_single_quotes");
}

# Check $option_double_quotes
my $DQUOTES_Q = "q";
my $DQUOTES_EM = "em";
my $DQUOTES_QEM = "qem";
$_ = $option_double_quotes;
if ($_ ne $DQUOTES_Q and $_ ne $DQUOTES_EM and $_ ne $DQUOTES_QEM) {
    &usage ("Invalid --double-quotes=$option_double_quotes");
}

# Check $option_xhtml
my $xhtml_slash = ($option_xhtml ? " /" : "");

# Check $option_strict_html
my $html_id = $option_strict_html ? "id" : "name";

# Check $option_output_type
my $TYPE_TEXT = "text";
my $TYPE_HTML = "html";
my $TYPE_FANCY_HTML = "fancy-html";
$_ = $option_output_type;
if ($_ ne $TYPE_TEXT and $_ ne $TYPE_HTML and $_ ne $TYPE_FANCY_HTML) {
    &usage ("Invalid --output-type=$option_output_type");
}

# Check $option_output
if (not defined $option_output) {
    if ($option_output_type eq $TYPE_TEXT) {
        $option_output = $ARGV[0].".txt";
    } elsif ($option_output_type eq $TYPE_HTML) {
        $option_output = $ARGV[0].".html";
    } else { # $option_output_type eq $TYPE_FANCY_HTML
        $option_output = $ARGV[0].".d";
    }
}
if (($option_output_type eq $TYPE_FANCY_HTML
     and not -d $option_output)) {
    mkdir $option_output
      or &usage ("Cannot make directory $option_output");
}

# Check FAQ file
open FAQ, "< $ARGV[0]"
  or &usage ("Cannot open file $ARGV[0]");


########################################################################
# Read the FAQ file

my $title_prefix   = "";           # The content of
my $charset        = "iso-8859-1"; # the directives
my $text_toc       = "Table Of Contents";
my $text_sections  = "Sections";
my $text_this_section      = "This section";
my $text_questions = "Questions";
my $text_qu        = "Q";
my $text_next      = "Next";
my $text_prev      = "Previous";
my $text_all       = "All sections";

my $linenum      = 0;           # Various counters when reading FAQ
my $partnum      = 0;
my $sectionnum   = 0;
my $questionnum  = 0;

my $title        = "";
my @content      = ();          # list of all paragraphs
my %ref          = ();          # $ref{"$ID"} = "$SECTION.$QUESTION"

my $para         = "";          # Current paragraph
my $lastpara     = "";          # Last paragraph

while (<FAQ>) {
    $linenum++;
    &process_line;
}
$partnum        = 0;
$_              = "";
&process_line;

sub process_line {

    # Process directives
    if (/^\?\{([-\w]+)=($RE_ELEMENT_CONTENT)\}/) {
        my $name        = $1;
        my $content     = $2;
        $content        =~ s/\\(.)/$1/g;
        if ($name eq "charset") {
            $charset            = $content;
        } elsif ($name eq "title-prefix") {
            $title_prefix       = $content;
        } elsif ($name eq "text-toc") {
            $text_toc           = $content;
        } elsif ($name eq "text-q") {
            $text_qu            = $content;
        } elsif ($name eq "text-next") {
            $text_next          = $content;
        } elsif ($name eq "text-prev") {
            $text_prev          = $content;
        } elsif ($name eq "text-this-section") {
            $text_this_section  = $content;
        } elsif ($name eq "text-sections") {
            $text_sections      = $content;
        } elsif ($name eq "text-questions") {
            $text_questions     = $content;
        } elsif ($name eq "text-all") {
            $text_all           = $content;
        } else {
            print STDERR "line $linenum: Unknown directive \`$name'\n";
        }
        s/^$RE_ELEMENT_CONTENT\}//; # cut the directive
    }

    return if $partnum > 2;

    # Line feed character `\f' separates parts.
    # Parts: 0=top, 1=questions, 2=bottom, rest is ignored
    if (/\f/) {
        s/\f//;                 # Remove it
        if (/\f/) {             # There is another one, this is error
            print STDERR "line $linenum: Illegal use of parts";
            s/\f//g;
        }
        $partnum++;
        push @content, "bottom:"        if $partnum == 2;
        return                          if $partnum > 2;
    }

    my $line    = $_;

    # Is it time to add the paragraph to @content?
    if ($para ne "" && (/^\s*$/ || /^[*-]\s/)) {

        # Process text markups

        # Escape special characters
        $para   =~ s/\\/\\\\/g; # `\'
        $para   =~ s/\{/\\\{/g; # `{'
        $para =~ s/\}/\\\}/g;   # `}'

        # `?keyword' (reference to question)
        $para   =~ s/(\W)\?(\w+)/$1\{ref=$2\}/g;

        # mail address
        $para   =~ s/(\W)(\w[\w.-]+\@[\w-]+(\.[\w-]+)+)/$1\{mail=$2\}/g;

        # http or URL address
        $para
            =~ s/(\W)((ftp|http):\/\/[^.)>\s]+(\.[^.)>\s]+)*)/$1\{href=$2\}/g;

        # ``TEXT'' that doesn't include `ANOTHER_TEXT'
        $para   =~ s/\`\`([^\`\']*)\'\'/\`\`\{q=$1\}\'\'/g;

        # `TEXT' that doesn't include ``ANOTHER_TEXT''
        $para   =~ s/\`(?!\{)([^\`\']*)\'/\`\{tt=$1\}\'/g;

        # ``TEXT'' where TEXT can include many `{tt=OTHER_TEXT}'
        $para   =~ s/\`\`([^\`\']*(?:\`\{tt=$RE_ELEMENT_CONTENT\}\'[^\`\']*)+)\'\'/\`\`\{q=$1\}\'\'/g;

        # `TEXT' where TEXT can include many ``{q=OTHER_TEXT}''
        $para   =~ s/\`(?!\{)([^\`\']*(?:\`\`\{q=$RE_ELEMENT_CONTENT\}\'\'[^\`\']*)+)\'/\`\{tt=$1\}\'/g;

        # Store the paragraph and prepare for the next one
        if ($title) {
            push @content, $para;
        } else {
            $title      = $para;
        }
        $lastpara       = $para;
        $para           = "";
    }
    $_  = $line;
    return if /^\s*$/;

    # Start of section (starts with ?)
    if ($para eq "" && /^\?\s/) {
        s/^\?\s+//;
        $para           = "section ".++$sectionnum.":".$_;
        $questionnum    = 0;
    }
    # Start of question (starts with ??[KEYWORD])
    elsif ($para eq "" && /^\?\?(\w*)/) {
        $questionnum++;
        if ($1 ne "") {
            $ref{$1}    = "$sectionnum.$questionnum";
        }
        s/^\?\?\w*\s*/\t/;
        $para   = "question $sectionnum.$questionnum:".$_;
    }
    # Start of list item (starts with * or -)
    elsif (/^[*-]\s/) {
        $para           = "item-start:".$_;
    }
    # Start of verbatim paragraph (starts with TAB or 3 spaces)
    elsif (/^(\t|\s{3,})/) {
        if ($para eq "") {
            if (substr($lastpara,0,4) eq "item") {
                $para   = "item-verbatim:";
            } else {
                $para   = "verbatim:";
            }
        }
        $para   .= $_;
    }
    # Start of normal paragraph (contains at least one nonspace char)
    elsif (/\S/) {
        if ($para eq "") {
            if (substr($lastpara,0,4) eq "item" && /^\s/) {
                $para   = "item-normal:";
            } else {
                $para   = "normal:";
            }
        }
        $para   .= $_;
    }
}

my $output_buffer = "";
my $item = 0;

sub do_output_text;
sub do_output_html;
sub do_output_html_fancy;

if ($option_output_type eq $TYPE_TEXT) {
    &do_output_text ();
} elsif ($option_output_type eq $TYPE_HTML) {
    &do_output_html ();
} else { # $option_output_type eq $TYPE_FANCY_HTML
    &do_output_html_fancy ();
}

########################################################################
# Output facilities

# output (string)
#   Outputs string to the buffer.
sub output {

    $output_buffer .= $_[0];
}

# output ()
#   Get the content of the buffer and empty it.
sub output_get {

    my $result = $output_buffer;
    $output_buffer = "";
    $result;
}


########################################################################
# Text output

sub output_text_text {
    $_          = " ".$_[0];    # Begin with space to easen REs.
    # Replace references
    while (/[^\\]\{ref=(\w+)\}/) {
        my $keyword     = $1;
        my $answer      = $ref{$keyword};
        s/([^\\])\{ref=$keyword\}/$1$text_qu$answer/g;
    }
    # Replace mail addresses and URLs
    s/([^\\])\{(mail|href)=($RE_ELEMENT_CONTENT)\}/$1$3/g;
    # Replace monospace text
    s/([^\\])\{tt=($RE_ELEMENT_CONTENT_2)\}/$1$2/g;
    # Replace quoted text
    s/([^\\])\{q=($RE_ELEMENT_CONTENT_2)\}/$1$2/g;
    # Common escaping
    s/\\\\/\\/g;
    s/\\\{/\{/g;
    s/\\\}/\}/g;
    # Cut the starting space that we inserted in the beginning of this
    # function
    s/^\s//s;

    output ($_);
}

sub output_text_para {

    $_          = $_[0];
    if (/^bottom:/) {
        output ("\f");
        return;
    }
    s/^.*?:/\n/s;
    output_text_text ($_);
}

sub output_text_centered {

    my $text    = $_[0];
    $text       =~ s/^.*?:([\s\S]*)$/$1/s;
    my @lines   = split /\n/, $text;
    foreach my $line (@lines) {
        $line           =~ s/^\s*(.*?)\s*$/$1/;
        output ((" " x (($TEXT_WIDTH - length $line) / 2))
                . $line . "\n");
    }
}

sub do_output_text {

    my $i;

    # Sections and questions are numbered.
    foreach (@content) {
        if (/^section/s || /^question/s) { s/^(.*?([\d.]*):)/$1$2 /s; }
    }

    # Title and top text.
    # Ends by reaching the first section.
    output_text_centered ($title);
    for ($i = 0; $i < $#content && $content[$i] !~ /^section/; $i++) {
        output_text_para ($content[$i]);
    }
    my $first_sections = $i;

    # Table Of Contents.  Only sections and questions.
    output ("\n\f\n");
    output_text_centered
      ("toc:". $text_toc ."\n". "-" x length $text_toc);
    for $i ($first_sections .. $#content) {
        $_ = $content[$i];
        if (/^section/s) {
	    output("\n");
            output_text_centered($_);
        } elsif (/^question/s) {
            output_text_para($_);
        }
    }

    # Body
    output ("\n\f\n");
    for $i ($first_sections .. $#content) {
        $_ = $content[$i];
        if (/^section/s) {
	    output("\n");
            output_text_centered ($_);
        } else {
            output_text_para ($_);
        }
    }

    open OUT, "> $option_output"
      or die "Cannot open $option_output for writing";
    print OUT output_get();
    close OUT;
}


########################################################################
# HTML output

sub output_html_internal_element {

    my ($begin, $end, $element_name, $to_string, $new_begin, $new_end) = @_;

    if (not defined $new_end) {	# by default do not change $begin and
	$new_end = $end;	# $end.
	if (not defined $new_begin) {
	    $new_begin = $begin
	}
    }

    $begin = quotemeta $begin;
    $end = quotemeta $end;
    $element_name = quotemeta $element_name;
    $new_begin = quotemeta $new_begin;
    $new_end = quotemeta $new_end;
    eval "s/([^\\\\])$begin\\{$element_name=($RE_ELEMENT_CONTENT_2)\\}$end/\$1$new_begin$to_string$new_end/g;\n";

    return $_;
}

sub output_html_element {

    my ($element, $content) = @_;

    local $_    = $content;     # Insert blank space to easen REs.
    s/^.*?:/ /s;                # Remove meta-info and add extra space

    # Special characters in HTML
    s/\&/\&amp;/g;
    s/\</\&lt;/g;
    s/\>/\&gt;/g;

    # References
    while (/[^\\]\{ref=(\w+)\}/) {
        my $keyword     = $1;
        my $answer      = $ref{$keyword};
        if ($answer eq "") {
            print STDERR "Reference to \`$keyword' is not valid";
        } else {
            if ($option_output_type eq $TYPE_FANCY_HTML) {
                $answer         =~
                  s/^(\d+)\.(\d+)$/<a href=\"\#q$1-$2\">$text_qu$1.$2<\/a>/;
            } else {
                $answer         =~
                  s/^(\d+)\.(\d+)$/<a href=\"q$1-$2.html\">$text_qu$1.$2<\/a>/;
            }
            s/([^\\])\{ref=$keyword\}/$1$answer/g;
        }
    }

    # URLs
    &output_html_internal_element ("", "", "href",
				   '\<a href=\"$2\"\>$2\<\/a\>');

    # Mail addresses
    &output_html_internal_element ("", "", "mail",
				   '\<a href=\"mailto:$2\"\>$2\<\/a\>');

    # Single quotes
    if ($option_single_quotes eq $SQUOTES_Q) {
	&output_html_internal_element ("`", "'", "tt", '$2');
    } elsif ($option_single_quotes eq $SQUOTES_TT) {
	&output_html_internal_element ("`", "'", "tt",
				       '\<tt\>$2\<\/tt\>', "", "");
    } else {			# eq $SQUOTES_QTT
	&output_html_internal_element ("`", "'", "tt",
				       '\<tt\>$2\<\/tt\>');
    }

    # Double quotes
    if ($option_double_quotes eq $DQUOTES_Q) {
	&output_html_internal_element ("``", "''", "q", '$2');
    } elsif ($option_double_quotes eq $DQUOTES_EM) {
	&output_html_internal_element ("``", "''", "q",
				       '\<em\>$2\<\/em\>', "", "");
    } else {			# eq $DQUOTES_QEM
	&output_html_internal_element ("``", "''", "q",
				       '\<em\>$2\<\/em\>');
    }

    # Common escaping
    s/\\\\/\\/g;
    s/\\\{/\{/g;
    s/\\\}/\}/g;
    # Remove the extra space
    s/^\s//s;

    # Print it as HTML wills
    output ("<$element>") if defined $element;
    output ($_);
    output ("</$element>\n\n") if defined $element;
}

sub output_html_toc {

    output ("<hr$xhtml_slash>\n\n");
    output ("<h4 $html_id='contents'>");
    output_html_element (undef, "toc:". $text_toc);
    output ("</h4>");

    my $first_section   = 1;
    output ("<ul>\n\n");
    if ($option_intro_in_toc) {
	$first_section      = 0;
	output (li() . strong() . "0 ");
	output ("<a href='#s0' $html_id='TOCs0'>");
	output_html_element (undef, $title);
	output ("</a></strong></li>\n\n<ul>");
    }
    foreach my $p (@content) {
        if ($p =~ /^section (\d+):\s*([\s\S]*?)\s*$/s) {
            output ("</ul></li>\n\n") unless $first_section;
            $first_section      = 0;
            output ("<li><strong>$1 ");
            output ("<a href='#s$1' $html_id='TOCs$1'>");
            output_html_element (undef, $p);
            output ("</a></strong>\n\n<ul>");
        } elsif ($p =~ /^question (\d+)\.(\d+):\s*([\s\S]*?)\s*$/s) {
            output ("<li>$1.$2 ");
            output ("<a href='#q$1-$2' id='TOCq$1-$2'>");
            output_html_element (undef, $p);
            output ("</a></li>\n");
        }
    }
    output ("</ul></li>\n\n</ul>\n\n") unless $first_section;

    output ("<hr$xhtml_slash>\n\n");
}

sub output_html_para {

    if ($item && ! /^item/s) {
        output ("</li>\n\n</ul>\n\n");
        $item   = 0;
    }

    if (/^item-start:/s) {
        if ($item) {
            output ("</li>\n\n");
        } else {
            output ("<ul>\n\n");
        }
        output ("<li>");
        $item   = 1;
        s/^(.*?:)[*-]/$1 /s;
        output_html_element ("p", $_);
    }
    elsif (/^normal:/s || /^item-normal:/s) {
        output_html_element ("p", $_);
    }
    elsif (/^verbatim:/s || /^item-verbatim:/s) {
        output_html_element ("pre", $_);
    }
    else {
        return 0;
    }

    return 1;
}

my $template;			# Holds HTML template

sub output_html_load_template {


    if ($option_html_template) {
	open TEMPLATE, "< $option_html_template";
	$template = "";
	while (<TEMPLATE>) {
	    $template .= $_;
	}
    } else {			# Use default template
	$template = <<EOT;
<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Strict//EN\"
	"http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
<meta http-equiv="Content-Type"
	content="text/html; charset=<gen-faq:charset/>" />
<title><gen-faq:title/></title>
</head>
<body>
<h2><gen-faq:title/></h2>
<gen-faq:all/>
</body>
</html>
EOT
    }

}

sub output_html_apply_template {

    my ($filename, $page_title, $all) = @_;
    $filename   = $option_output if not defined $filename;
    $page_title = $title         if not defined $page_title;
    $all        = output_get()   if not defined $all;

    my $save_output_buffer = output_get();
    output_text_para ($page_title);
    my $fine_title = output_get();
    $output_buffer = $save_output_buffer;

    $_ = $template;
    my $re_charset = quotemeta $charset;
    s/\<gen-faq:charset\s*\/\>/$charset/g;
    my $re_title = quotemeta $fine_title;
    s/\<gen-faq:title\s*\/\>/$fine_title/g;
    my $re_all = quotemeta $all;
    s/\<gen-faq:all\s*\/\>/$all/g;

    open OUT, "> $filename"
      or die "Cannot open \`$filename\' for writing";
    print OUT $_;
    close OUT;
}

sub output_html_title {

    my $custom_title = $_[0];
    $custom_title = $title if not defined $custom_title;

    print STDERR output_get ();
    die "Internal error: output_get() ne '' before output_html_title()"
      if output_get () ne "";

    output_text_para ($custom_title);
    &output_html_element (undef, "x:" . output_get ());
    $title = output_get ();
}

sub do_output_html {

    output_html_load_template ();

    output_html_title ();

    my $first_section   = 1;
    $item               = 0;
    foreach (@content) {
        if (not output_html_para ()) {
            if (/^section (\d+):/s) {
		if ($first_section) {
		    my $intro = output_get ();
		    if (not $option_intro_in_toc) {
			output ($intro);
		    }
		    output_html_toc ();
		    if ($option_intro_in_toc) {
			output ("<h3>0 <a href=\"#TOCs0\" $html_id=\"s0\">");
			output_html_element (undef, $title);
			output ("</a></h3>\n" . $intro);
		    }
		    $first_section  = 0;
		}
                output ("<h3>$1 <a href=\"#TOCs$1\" $html_id=\"s$1\">");
                output_html_element (undef, $_);
                output ("</a></h3>\n\n");
            }
            elsif (/^question/s) {
                /^question (\d+)\.(\d+):/s;
                output ("<h4>$1.$2 "
                  ."<a href=\"#TOCq$1-$2\" $html_id=\"q$1-$2\">");
                output_html_element (undef, $_);
                output ("</a></h4>");
            }
            elsif (/^bottom/s) {
                output ("<hr$xhtml_slash>\n\n");
            } else {
                &usage ("Internal error: unknown mark up \`$_'\n");
                output_html_element ("unknown", $_);
            }
        }
    }
    $_ = "normal:", output_html_para ();

    output_html_apply_template ()
}


########################################################################
# XHTML fancy output

sub output_html_fancy_sections {

    my $element = $_[0];
    my $current = $_[1];
    my $current_linked  = $_[2];

    # List Of Sections
    if ($element eq "p") {
        output ("<p>");
        output_html_element ("strong", "los:$text_sections");
        output ("</p>\n\n<blockquote>\n\n");
    } else {
        output ("<p><small><a href=\"index.html\">");
        output_html_element ("strong", "los:$text_sections");
        output ("</a></small></p>\n\n<ul>\n\n");
    }
    foreach (@content) {
        if (/^section (\d+):([\s\S]*)$/s) {
            output ("<li>") if $element ne "p";
            my $href = "s$1.html";
            my $elem = $element;
            if (defined $current && $_ eq $current) {
                output ("<$element>");
                $elem   = "strong";
                undef $href if not defined $current_linked;
            }
	    output ("<$elem>$1 ");
	    output ("<a href='$href'>") if defined $href;
            output_html_element (undef, $_);
	    output ("</a>") if defined $href;
	    output ("</$elem>");
            output ("</$element>") if not defined $href;
            output ("</li>") if $element ne "p";
        }
    }
    output ("</blockquote>\n\n")        if $element eq "p";
    output ("</ul>\n\n")                if $element ne "p";
}

sub do_output_html_fancy {

    output_html_load_template ();

    my $dir = $option_output;
    mkdir $dir;
    die "Non-existent directory \`$dir'\n" if not -d $dir;

    # Top page
    {
        # Top text
        $item   = 0;
        foreach (@content) {
            last if (/^section/s);
            output_html_para ();
        }
        $_ = "";
        output_html_para ();
        output_html_fancy_sections ("p");
        $item                   = 0;
        my $it_is_bottom        = 0;
        foreach (@content) {
            output_html_para () if $it_is_bottom;
            $it_is_bottom       = 1 if /^bottom/s;
        }
	$_ = "normal:",	output_html_para ();
	output_html_apply_template ("$dir/index.html");
    }

    # Sections
    {
        foreach my $section (@content) {
            if ($section =~ /^section (\d+):([\s\S]*)/s) {
		my $section_filename = "$dir/s$1.html";
		my $section_title    = "title:$1 $2";
                output ("<hr$xhtml_slash>\n\n");
                output_html_fancy_sections ("small", $section);
                output ("<hr$xhtml_slash><p>");
                output_html_element ("strong", "q:$text_questions");
                output ("</p>\n\n<ul>\n\n");
                my $started     = 0;
                foreach (@content) {
                    last if $started and /^section/s;
                    $started = 1 if not $started and $_ eq $section;
		    if ($started and /^question (\d+)\.(\d+):([\s\S]*)/s) {
			output ("<li><p>$1.$2 <a href='q$1-$2.html'>");
			output_html_element (undef, $_);
			output ("</a></p></li>\n");
		    }
                }
                output ("</ul>\n\n<hr$xhtml_slash>\n\n");
                output_html_fancy_sections ("small", $section);
                output ("<hr$xhtml_slash>\n\n");
		output_html_apply_template ($section_filename, $section_title);
	    }
	}
    }

    # Questions
    {
        my $section             = ":"; # no section
        for my $i (0 .. $#content) {
            if ($content[$i] =~ /^section/s) {
                $section        = $content[$i];
            } elsif ($content[$i] =~ /^question (\d+).(\d+):([\s\S]*)$/s) {
		my $question_filename = "$dir/q$1-$2.html";
                my $question_title    = "title:$text_qu$1.$2 $3";
                my $sectionnum  = $1;
                my $questionnum = $2;
                my $title       = $3;
                my $nav         = "no navigation bar";
                # Navigation bar
                {
                    output ("<div class='nav' style='text-align: center'>");
                    # Is there "next"?
                    my $j       = $i+1;
                    while ($j <= $#content &&
                           $content[$j] !~ /^question (\d+)\.(\d+)/s) {
                        $j++;
                    }
                    if ($j <= $#content) {
                        $content[$j] =~ /^question (\d+)\.(\d+)/s;
                        output ("[<a href='q$1-$2.html'>$text_next</a>] ");
                    }
                    # "prev"?
                    $j          = $i-1;
                    while ($j >= 0 &&
                           $content[$j] !~ /^question (\d+)\.(\d+)/s) {
                        $j--;
                    }
                    if ($j >= 0) {
                        $content[$j] =~ /^question (\d+)\.(\d+):/s;
                        output ("[<a href='q$1-$2.html'>$text_prev</a>] ");
                    }
                    output ("[<a href='s$sectionnum.html'>$text_this_section</a>] ");
                    output ("[<a href='index.html'>$text_all</a>]");
                    output ("</div>\n\n");
		    $nav = output_get ();
                }
		# Top navigation
                output ("<hr$xhtml_slash>\n\n");
		output ($nav);
                output ("<hr$xhtml_slash>\n\n");
                # Header -- title is inserted by the template
                #output_html_element ("h3",
                #                      "x:$text_qu$sectionnum.$questionnum $title");
                # Body
                {
                    my $j       = $i+1;
                    $item       = 0;
                    while ($j <= $#content
                           && $content[$j] !~ /^question/s
                           && $content[$j] !~ /^section/s
                           && $content[$j] !~ /^bottom/s) {
                        $_      = $content[$j];
                        output_html_para ();
                        $j++;
                    }
                    output ("</ul>\n\n") if $item;
                }
                # Bottom navigation
                output ("<hr$xhtml_slash>");
		output ($nav);
                output ("<hr$xhtml_slash>\n\n");
		output_html_apply_template ($question_filename, $question_title);
            } # if question
        } # for each paragraph
    } # output questions
} # do_output_html_fancy
